from PIL import ImageFont, ImageDraw, Image
from skimage.metrics import structural_similarity as ssim
import numpy as np
import logging
import os

from koba.core import font


FONT_SIZE = 20

font_path = font.get_monospace_font()
font_cache = {}
char_cache = {}

def get_font(char):
    if char not in font_cache.keys():
        if font_path and font.check_support(char, font_path):
            font_cache[char] = ImageFont.truetype(font_path, FONT_SIZE)
        else:
            new_font = font.get_supported_font(char)
            if new_font:
                logging.debug(f"Found font that supports {char}.")
                font_cache[char] = ImageFont.truetype(new_font, FONT_SIZE)
            else:
                return None
    return font_cache[char]

# from UCYT5040/lectrick
def crop_image(image):
    image_array = np.array(image)
    non_white_mask = image_array < 0
    non_white_pixels = np.argwhere(non_white_mask)
    if non_white_pixels.size == 0:
        return image
    top, left = non_white_pixels.min(axis=0)
    bottom, right = non_white_pixels.max(axis=0) + 1  # +1 to include the last pixel
    return image.crop((left, top, right, bottom))

def get_char(char, width, height, save=False):
    char_arr = char_cache.get((char, width, height))
    if char_arr is not None:
        return char_arr
    else:
        font = get_font(char)
        if not font:
            logging.critical("Some chars are not supported by any of your installed fonts. Please specify a custom font with the --font option.")
            raise FileNotFoundError
        bbox = font.getbbox(char)
        
        if not (bbox and bbox[2] > bbox[0] and bbox[3] > bbox[1]):
            return None
        
        char_width, char_height = bbox[2] - bbox[0], bbox[3] - bbox[1]
        
        char_image = Image.new('L', (char_width, char_height), 0)
        draw = ImageDraw.Draw(char_image)
        draw.text((-bbox[0], -bbox[1]), char, font=font, fill=255)
                
        char_image = crop_image(char_image)
        
        char_image = char_image.resize((width, height), Image.Resampling.BILINEAR)
        
        if save:
            os.makedirs("chars", exist_ok=True)
            char_code = ord(char)
            char_image.save(f"chars/{char_code}.png")
        
        char_arr = np.array(char_image)
            
        char_cache[(char, width, height)] = char_arr
        return char_arr

def pre_render_characters(characters, sizes, save_chars, progress_callback=None):
    for size in sizes:
        width, height = size
        for char in characters:
            get_char(char, width, height, save_chars)
            if progress_callback:
                progress_callback()

def compare_character(char, block_arr, save_chars, engine):
    height, width = block_arr.shape
    
    char_arr = get_char(char, width, height, save=save_chars)
    if char_arr is None:
        return 0.0
    
    if char_arr.shape != block_arr.shape:
        char_img = Image.fromarray(char_arr)
        char_img = char_img.resize((width, height), Image.Resampling.BILINEAR)
        char_arr = np.array(char_img)
    
    # compare arrays
    if engine == "brightness":
        avg_char = np.mean(char_arr)
        avg_block = np.mean(block_arr)
        score = abs(avg_char - avg_block)
        similarity = 1.0 - (score / 255.0)
    elif engine == "ssim":
        min_side = min(block_arr.shape)
        win_size = min(7, min_side)
        if win_size % 2 == 0:
            win_size -= 1
        if win_size < 3:
            win_size = 3
        
        try:
            score = ssim(block_arr, char_arr, data_range=255, win_size=win_size)
            if isinstance(score, tuple):
                score = score[0]
            similarity = max(0.0, float(score))
        except ValueError:
            # This can happen if the window size is larger than the image
            similarity = 0.0
            
    elif engine == "diff":
        pixel_diff = np.sum(np.abs(block_arr.astype(np.int16) - char_arr.astype(np.int16)))
        max_diff = float(width * height * 255)
        if max_diff == 0:
            return 1.0
        dissimilarity = pixel_diff / max_diff
        similarity = 1.0 - dissimilarity
    elif engine == "mse":
        mse = np.mean((block_arr.astype(np.int32) - char_arr.astype(np.int32)) ** 2)
        max_mse = 255.0 ** 2
        similarity = 1.0 - (mse / max_mse) if max_mse > 0 else 1.0
    elif engine == "ncc":
        a = block_arr.astype(np.float64)
        b = char_arr.astype(np.float64)
        
        std_a = a.std()
        std_b = b.std()

        if std_a == 0 and std_b == 0:
            similarity = 1.0 if a.mean() == b.mean() else 0.0
        elif std_a == 0 or std_b == 0:
            similarity = 0.0
        else:
            a = (a - a.mean()) / std_a
            b = (b - b.mean()) / std_b
            ncc = np.mean(a * b)
            similarity = (ncc + 1) / 2
    elif engine == "hist":
        hist_a, _ = np.histogram(block_arr, bins=256, range=(0,255), density=True)
        hist_b, _ = np.histogram(char_arr, bins=256, range=(0,255), density=True)
        similarity = np.minimum(hist_a, hist_b).sum()
    elif engine == "cosine":
        a = block_arr.flatten().astype(np.float64)
        b = char_arr.flatten().astype(np.float64)
        
        norm_a = np.linalg.norm(a)
        norm_b = np.linalg.norm(b)

        if norm_a == 0 and norm_b == 0: # Both are black
            similarity = 1.0
        elif norm_a == 0 or norm_b == 0: # One is black
            similarity = 0.0
        else:
            cos_sim = np.dot(a, b) / (norm_a * norm_b)
            similarity = (cos_sim + 1) / 2
    else:
        raise ValueError(f"Invalid engine specified: {engine}")
        
    return similarity
    
def get_character(img_arr, characters, engine, save_chars):
    max_similarity = float("-inf")
    best_match = " "
    for character in characters:
        similarity = compare_character(character, img_arr, save_chars, engine)
        if similarity > max_similarity:
            max_similarity = similarity
            best_match = character
    
    return best_match


# TODO: dont depend on matplotlib

import matplotlib.font_manager as fm
from fontTools.ttLib import TTFont

# cache for font paths and codepoints
_font_paths = None
_font_codepoints = {}

def get_all_font_paths():
    global _font_paths
    if _font_paths is None:
        _font_paths = [f.fname for f in fm.fontManager.ttflist]
    return _font_paths

def get_monospace_font():
    for font in fm.fontManager.ttflist:
        if "dejavusansmono" in font.name.replace(" ", "").lower():
            return font.fname
    return None

def get_supported_font(char):
    codepoint = ord(char)
    for font_path in get_all_font_paths():
        if font_supports_codepoint(font_path, codepoint):
            return font_path
    return False

def font_supports_codepoint(font_path, codepoint):
    if font_path not in _font_codepoints:
        try:
            font = TTFont(font_path)
            cps = set()
            for cmap in font['cmap'].tables:
                if cmap.isUnicode():
                    cps.update(cmap.cmap.keys())
            _font_codepoints[font_path] = cps
        except Exception:
            _font_codepoints[font_path] = set()
    return codepoint in _font_codepoints[font_path]

def check_support(char, font_path):
    return font_supports_codepoint(font_path, ord(char))


from koba.core import charsets

def test_get_range_ascii():
    chars = charsets.get_range(32, 126)
    assert len(chars) == 95
    assert ' ' in chars
    assert '~' in chars
    assert '\n' not in chars

def test_get_range_single_char():
    chars = charsets.get_range(65, 65)
    assert chars == ['A']

def test_get_range_empty():
    chars = charsets.get_range(100, 90)
    assert len(chars) == 0

def test_get_amount():
    chars = charsets.get_amount(10)
    assert len(chars) == 10
    assert all(c.isprintable() for c in chars)



from click.testing import CliRunner
from koba.main import main
from PIL import Image
import os
import pytest

def test_cli_runs_successfully(monkeypatch):
    # mock get_terminal_size to avoid errors
    monkeypatch.setattr(os, 'get_terminal_size', lambda: os.terminal_size((80, 24)))

    runner = CliRunner()
    img = Image.new('L', (20, 20), color='black')
    test_image_path = 'test_image.png'
    img.save(test_image_path)

    result = runner.invoke(main, [test_image_path, '--engine', 'brightness'])

    os.remove(test_image_path)

    assert result.exit_code == 0
    assert result.output.strip() != ""

def test_cli_file_not_found():
    runner = CliRunner()
    result = runner.invoke(main, ['non_existent_file.png'])

    assert result.exit_code != 0
    assert "does not exist" in result.output

def test_cli_bad_char_range(caplog):
    runner = CliRunner()
    img = Image.new('L', (10, 10), color='black')
    test_image_path = 'test_image.png'
    img.save(test_image_path)

    result = runner.invoke(main, [test_image_path, '--char-range', 'invalid-range'])

    os.remove(test_image_path)

    assert result.exit_code == 1
    assert isinstance(result.exception, SystemExit)
    assert "Wrong format for char-range." in caplog.text


from koba.core import font

def test_get_supported_font():
    font_path = font.get_supported_font("a")
    if font_path:
        assert True
    else:
        assert False
        
def test_get_monospace_font():
    font_path = font.get_monospace_font()
    if font_path:
        assert True
    else:
        assert False


import os
import sys
import math
import time
import logging
import concurrent.futures

import click
import numpy as np
from tqdm import tqdm
from PIL import Image, ImageOps, UnidentifiedImageError

from koba import __version__
from koba.core import charsets, unify

def process_block(args):
    block, characters, engine, save_chars = args
    return unify.get_character(block, characters, save_chars, engine)

def calculate_block_sizes(width, height, char_aspect, scale):
    terminal_width = os.get_terminal_size().columns
    chars_width = terminal_width
    
    min_block_width = 10 / char_aspect
    min_block_height = 10

    scale = min(scale, 1.0)
    
    max_chars_width_by_block_width = int(width // min_block_width)
    max_chars_width_by_block_height = int(height // min_block_height * char_aspect)

    max_chars_width = min(max_chars_width_by_block_width, max_chars_width_by_block_height)

    chars_width = min(int(terminal_width * scale), terminal_width, max_chars_width)

    chars_width = max(chars_width, 1)

    chars_height = math.ceil((height * chars_width / width) / char_aspect)
    
    block_widths = [width // chars_width] * chars_width
    for i in range(width % chars_width):
        block_widths[i] += 1

    block_heights = [height // chars_height] * chars_height
    for i in range(height % chars_height):
        block_heights[i] += 1
        
    return block_widths, block_heights, chars_width

def process(img, char_aspect, scale, engine, color, invert, stretch_contrast, save_blocks, start_char, end_char, save_chars, font, single_threaded, show_progress=False):
    if color:
        img_color = img.convert("RGB")
        img_arr_color = np.array(img_color)
    
    img = img.convert("L")
    if invert:
        img = ImageOps.invert(img)
    if stretch_contrast:
        img = ImageOps.autocontrast(img)
    
    img_arr = np.array(img)
    height, width = img_arr.shape[:2]
    logging.debug(f"Image is {width}x{height} pixels.")
    
    block_widths, block_heights, chars_width = calculate_block_sizes(width, height, char_aspect, scale)

    logging.debug(f"Image will be {chars_width}x{len(block_heights)} chars.")
    logging.debug(f"Block widths: {block_widths[:5]}... (total {len(block_widths)})")
    logging.debug(f"Block heights: {block_heights[:5]}... (total {len(block_heights)})")

    if min(block_heights + block_widths) <= 7 and str(engine).lower() == "ssim":
        logging.critical("Image blocks are too small for SSIM. Please use another engine.")
        sys.exit(1)

    # splitting into blocks with variable sizes
    block_colors = []
    blocks = []
    y = 0
    for bh in block_heights:
        x = 0
        for bw in block_widths:
            block = img_arr[y:y+bh, x:x+bw]                
            blocks.append(block)
            x += bw
        y += bh
    
    if color:
        y = 0
        for bh in block_heights:
            x = 0
            for bw in block_widths:
                block = img_arr_color[y:y+bh, x:x+bw]
                block_color = block.mean(axis=(0, 1)).astype(int)            
                block_colors.append(block_color)
                x += bw
            y += bh
    
    if color:
        if len(block_colors) != len(blocks):
            logging.critical("Block colors don't match the blocks. Falling back to grayscale.")
            color = False
        else:
            logging.debug(f"Block colors: {block_colors[:3]} ...")
    
    logging.debug(f"{len(blocks)} blocks created.")
    
    if save_blocks:
        logging.info("Saving image blocks...")
        blocks_dir = "blocks"
        os.makedirs(blocks_dir, exist_ok=True)
        for idx, block in enumerate(blocks):
            img_block = Image.fromarray(block)
            img_block.save(os.path.join(blocks_dir, f"{idx:04d}.png"))
    
    print_chars = ""
    
    if start_char == end_char:
        all_chars = chr(start_char) * len(blocks)
    else:
        characters = charsets.get_range(start_char, end_char)
        all_chars = ""

        # prepare arguments
        unique_blocks = {block.tobytes(): block for block in blocks}
        unique_block_args = [(block, characters, engine.lower(), save_chars) for block in unique_blocks.values()]

        if font:
            unify.font_path = font

        block_results = {}
        if not single_threaded:
            with concurrent.futures.ProcessPoolExecutor() as executor:
                future_to_block = {
                    executor.submit(unify.get_character, *arg): arg[0] for arg in unique_block_args
                }
                try:
                    for future in tqdm(
                        concurrent.futures.as_completed(future_to_block),
                        total=len(future_to_block),
                        desc="Processing unique blocks",
                        disable=not show_progress
                    ):
                        block = future_to_block[future]
                        block_results[block.tobytes()] = future.result()
                except ValueError as e:
                    logging.critical(str(e))
                    sys.exit(1)
        else:
            for args in tqdm(
                unique_block_args,
                desc="Processing unique blocks (single-threaded)",
                disable=not show_progress
            ):
                block, characters, engine, save_chars = args
                result = unify.get_character(block, characters, save_chars, engine)
                block_results[block.tobytes()] = result

        results = [block_results[block.tobytes()] for block in blocks]
        
        all_chars = "".join(results)
    
    lines = [all_chars[i:i+chars_width] for i in range(0, len(all_chars), chars_width)]
    flat_chars = "".join(lines)
    logging.debug(f"Char count: {len(flat_chars)}")
    logging.debug(f"Colored blocks: {len(block_colors)}")

    color_idx = 0
    for c in '\n'.join(lines):
        if c == '\n':
            print_chars += '\n'
        elif color_idx < len(block_colors):
            r, g, b = block_colors[color_idx]
            print_chars += f"\033[38;2;{r};{g};{b}m{c}\033[0m"
            color_idx += 1
        else:
            print_chars += c
            
    return print_chars
